/* simpleindex.cc: Index each paragraph in a textfile as a document
 *
 * ----START-LICENCE----
 * Copyright 1999,2000,2001 BrightStation PLC
 * Copyright 2002 Ananova Ltd
 * Copyright 2002 Olly Betts
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 * -----END-LICENCE-----
 */

#include <om/om.h>

#include <algorithm>
#include <iostream>
#include <fstream>
#include <string>
#include <vector>

using namespace std;

#include <ctype.h>

// Put a limit on the size of terms to help prevent the index being bloated
// by useless junk terms
static const unsigned int MAX_PROB_TERM_LENGTH = 64;

class ExpandDeciderOmega : public OmExpandDecider {
    private:
	OmDatabase db;
    public:
	ExpandDeciderOmega(const OmDatabase &db_) : db(db_) { }
	int operator()(const om_termname & tname) const {
	    // only suggest 4 or more letter words for now to
	    // avoid italian problems FIXME: fix this at index time
	    if (tname.length() <= 3) return false;

	    // Ignore a term that only occurs once (a hapax) since it's not
	    // useful for finding related documents - it only occurs in a
	    // document that's already been marked as relevant
	    if (db.get_termfreq(tname) <= 1) return false;

	    // Raw terms are OK, otherwise avoid terms with a prefix or with a space in
	    if (tname[0] == 'R') return true;
	    if (isupper(tname[0]) || tname.find(' ') != string::npos)
		return false;

	    return true;
	}
};

inline static bool
p_alnum(unsigned int c)
{
    return isalnum(c);
}

inline static bool
p_notalnum(unsigned int c)
{
    return !isalnum(c);
}

inline static bool
p_notplusminus(unsigned int c)
{
    return c != '+' && c != '-';
}

static void
lowercase_term(om_termname &term)
{
    om_termname::iterator i = term.begin();
    while (i != term.end()) {
	*i = tolower(*i);
	i++;
    }
}

int main(int argc, char **argv)
{
    if (argc != 3 && argc != 4) {
	cout << "usage: " << argv[0] << " DB DOCID [LENGTH]" << endl;
	exit(1);
    }

    om_doccount requested = 10;
    if (argc == 4) requested = atoi(argv[3]);

    vector<string> terms;
    string file;
    try {
	// Open the database
	OmSettings settings;
	settings.set("backend", "auto");
	settings.set("auto_dir", argv[1]);
	OmDatabase db(settings);

	OmEnquire enquire(db);

	OmRSet rset;
	om_docid did = atoi(argv[2]);
	rset.add_document(did);

	ExpandDeciderOmega decider(db);
	OmESet eset = enquire.get_eset(10, rset, 0, &decider);

	copy(eset.begin(), eset.end(), back_inserter(terms));

	string s = db.get_document(did).get_data();
	string::size_type l = s.find('\n');
	file = string(".") + s.substr(4, l - 4);
    } catch (const OmError &error) {
	cerr << "Exception: "  << error.get_msg() << endl;
	exit(1);
    }

    cout << "'" << file << "':\n";
    ifstream in(file.c_str());

    // Open temporary inmemory database to index sentences
    OmWritableDatabase *temp_db;
    try {
	// Create a database
	OmSettings settings;
	settings.set("backend", "inmemory");
	temp_db = new OmWritableDatabase(settings);
    }
    catch (const OmError &error) {
	cerr << "Exception: "  << error.get_msg() << endl;
	exit(1);
    }
    
    OmStem stemmer("english");
    string sentence;
    while (!in.eof()) {
	string word;
	in >> word;
	if (!word.empty()) {
	    if (!sentence.empty()) sentence += ' ';
	    sentence += word;
	    switch (word[word.size() - 1]) {
		case '\n':
		    if (in.eof()) break;
		    char ch;
		    in.get(ch);
		    if (ch != '\n') {
			in.unget();
			break;
		    }
		case '.': case '?': case '!': case ':': case ';':
		    try {
			OmDocument doc;
			doc.set_data(sentence);

			om_termcount pos = 0;
			string::iterator i, j = sentence.begin(), k;
			while ((i = find_if(j, sentence.end(), p_alnum)) != sentence.end())
			{
			    j = find_if(i, sentence.end(), p_notalnum);
			    k = find_if(j, sentence.end(), p_notplusminus);
			    if (k == sentence.end() || !isalnum(*k)) j = k;
			    string::size_type len = j - i;
			    if (len <= MAX_PROB_TERM_LENGTH) {
				om_termname term = sentence.substr(i - sentence.begin(),
					len);
				lowercase_term(term);
				term = stemmer.stem_word(term);
				doc.add_posting(term, pos++);
			    }
			}

			// Add the document to the database
			temp_db->add_document(doc);
		    } catch (const OmError &error) {
			cerr << "Exception: "  << error.get_msg() << endl;
		    }
		    sentence = "";
		    break;
	    }
	}
    }

    OmEnquire enquire(*temp_db);
    OmQuery query(OmQuery::OP_OR, terms.begin(), terms.end());
    cout << query.get_description() << "\n===" << endl;
    enquire.set_query(query);
    OmMSet mset = enquire.get_mset(0, requested);
    vector<om_docid> sorted(mset.begin(), mset.end());
    sort(sorted.begin(), sorted.end());
    // Work out how many extra we need
    if (sorted.size() < requested) {
	requested -= sorted.size();
    } else {
	requested = 0;
    }
    vector<om_docid>::const_iterator i;
    om_docid prev = 0;
    for (i = sorted.begin(); i != sorted.end(); ++i) {
	om_docid did = *i;
	if (prev + 1 < did) {
	    while (requested && ++prev < did) {
		cout << temp_db->get_document(prev).get_data();
		--requested;
	    }
	    if (prev + 1 < did) cout << " ...";
	}
	if (did > 1) cout << ' ';
	cout << temp_db->get_document(did).get_data();
	prev = did;
    }
    if (requested) {
	// Docs must be contiguous so far, so no ...
	while (requested && prev < temp_db->get_doccount()) {
	    if (prev > 0) cout << ' ';
	    cout << temp_db->get_document(++prev).get_data();
	    --requested;
	}
    }
    if (prev < temp_db->get_doccount()) cout << " ...";
    cout << endl;
    
    delete temp_db;
    return 0;
}
